@@ -13,7 +13,7 @@
 //! }
 //!
 //! let bench = Bench::new();
-//! let options = Options::default();
+//! let mut options = Options::default();
 //! let res = bench.run(&options, || test_function());
 //! println!("result: {}", res);
 //! ```
@@ -34,6 +34,29 @@
 //! let throughput = res.throughput(m.len() as _);
 //! println!("throughput: {}", throughput);
 //! ```
+//! Options:
+//!
+//! ```rust
+//! pub struct Options {
+//!     /// Number of iterations to perform.
+//!     pub iterations: u64,
+//!     /// Number of warm-up iterations to perform.
+//!     pub warmup_iterations: u64,
+//!     /// Minimum number of samples to collect.
+//!     pub min_samples: usize,
+//!     /// Maximum number of samples to collect.
+//!     pub max_samples: usize,
+//!     /// Maximum RSD to tolerate (in 0...100).
+//!     pub max_rsd: f64,
+//!     /// Maximum benchmark duration time.
+//!     pub max_duration: Option<std::time::Duration>,
+//!     /// Verbose output
+//!     pub verbose: bool,
+//! }
+//! ```
+//!
+//! Benchmark results can be made verbose by setting `verbose` to `true` in the
+//! `Options` struct, or by defining a `BENCHMARK_VERBOSE` environment variable.
 
 use precision::*;
 use std::fmt::{self, Debug, Display, Formatter};
@@ -41,24 +64,42 @@ use std::mem;
 use std::ops::Add;
 use std::ptr;
 use std::rc::Rc;
+use std::time::Duration;
 
 /// Options.
 #[derive(Clone, Debug)]
 pub struct Options {
     /// Number of iterations to perform.
     pub iterations: u64,
+    /// Number of warm-up iterations to perform.
+    pub warmup_iterations: u64,
+    /// Minimum number of samples to collect.
+    pub min_samples: usize,
     /// Maximum number of samples to collect.
     pub max_samples: usize,
-    /// Maximum RSD to tolerate (in 0...100)
+    /// Maximum RSD to tolerate (in 0...100).
     pub max_rsd: f64,
+    /// Maximum benchmark duration time.
+    pub max_duration: Option<Duration>,
+    /// Verbose output
+    pub verbose: bool,
 }
 
 impl Default for Options {
     fn default() -> Self {
+        let mut verbose = false;
+        std::env::var("BENCHMARK_VERBOSE")
+            .map(|_| verbose = true)
+            .ok();
+
         Self {
             iterations: 1,
-            max_samples: 3,
+            warmup_iterations: 0,
+            min_samples: 3,
+            max_samples: 5,
             max_rsd: 5.0,
+            verbose,
+            max_duration: None,
         }
     }
 }
@@ -177,6 +218,21 @@ impl Throughput {
     pub fn as_gb(&self) -> f64 {
         self.volume * 1_000_000_000f64 / (self.result.as_ns() as f64) / (1000.0 * 1000.0 * 1000.0)
     }
+
+    /// The throughput in kilobits.
+    pub fn as_kb8(&self) -> f64 {
+        self.volume * 8_000_000_000f64 / (self.result.as_ns() as f64) / 1000.0
+    }
+
+    /// The throughput in megabits.
+    pub fn as_mb8(&self) -> f64 {
+        self.volume * 8_000_000_000f64 / (self.result.as_ns() as f64) / (1000.0 * 1000.0)
+    }
+
+    /// The throughput in gigabits.
+    pub fn as_gb8(&self) -> f64 {
+        self.volume * 8_000_000_000f64 / (self.result.as_ns() as f64) / (1000.0 * 1000.0 * 1000.0)
+    }
 }
 
 impl Display for Throughput {
@@ -209,14 +265,14 @@ impl Bench {
         Bench { precision }
     }
 
-    pub fn run_once<F>(&self, options: Rc<Options>, f: &mut F) -> BenchResult
+    fn run_once<F, G>(&self, options: Rc<Options>, f: &mut F) -> BenchResult
     where
-        F: FnMut(),
+        F: FnMut() -> G,
     {
         let iterations = options.iterations;
         let start = self.precision.now();
         for _ in 0..iterations {
-            f();
+            black_box(f());
         }
         let elapsed = self.precision.now() - start;
         BenchResult {
@@ -227,17 +283,35 @@ impl Bench {
     }
 
     /// Run a single test.
-    pub fn run<F>(&self, options: &Options, mut f: F) -> BenchResult
+    pub fn run<F, G>(&self, options: &Options, mut f: F) -> BenchResult
     where
-        F: FnMut(),
+        F: FnMut() -> G,
     {
         let options = Rc::new(options.clone());
         let max_samples = std::cmp::max(1, options.max_samples);
+        let verbose = options.verbose;
+
+        if verbose {
+            println!("Starting a new benchmark.");
+            if options.warmup_iterations > 0 {
+                println!("Warming up for {} iterations.", options.warmup_iterations);
+            }
+        }
+        for _ in 0..options.warmup_iterations {
+            black_box(f());
+        }
         let mut results = Vec::with_capacity(max_samples as usize);
-        for _ in 0..max_samples {
+        let start = self.precision.now();
+        for i in 1..=max_samples {
+            if verbose {
+                println!("Running iteration {}.", i);
+            }
             let result = self.run_once(options.clone(), &mut f);
             results.push(result);
             if results.len() <= 1 {
+                if verbose {
+                    println!("Iteration {}: {}", i, results.last().unwrap());
+                }
                 continue;
             }
             let mean = results.iter().map(|r| r.as_secs_f64()).sum::<f64>() / results.len() as f64;
@@ -248,11 +322,31 @@ impl Bench {
                 / (results.len() - 1) as f64)
                 .sqrt();
             let rsd = std_dev * 100.0 / mean;
-            if rsd < options.max_rsd {
+            if verbose {
+                println!("Iteration {}: {:.2}s ± {:.2}%", i, mean, rsd);
+            }
+            if i >= options.min_samples && rsd < options.max_rsd {
+                if verbose {
+                    println!("Enough samples have been collected.");
+                }
                 break;
             }
+            if let Some(max_duration) = options.max_duration {
+                let elapsed =
+                    Duration::from_secs((self.precision.now() - start).as_secs(&self.precision));
+                if elapsed >= max_duration {
+                    if verbose {
+                        println!("Timeout.");
+                    }
+                    break;
+                }
+            }
         }
-        results.into_iter().min_by_key(|r| r.as_ns()).unwrap()
+        let result = results.into_iter().min_by_key(|r| r.as_ns()).unwrap();
+        if verbose {
+            println!("Result: {}", result);
+        }
+        result
     }
 }
 
